/*
    University timetable main file.
 */

import java.io.*;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

public class timetable {

    // Declaring all the input data (Time Slots, Rooms, Commitments and Modules)
    private List<String[]> timeSlots = new ArrayList<>();
    private List<String> rooms = new ArrayList<>();
    private List<String[]> commitments = new ArrayList<>();
    private HashMap<String, List<String>> modules = new HashMap<>();
    private Runtime runtime;

    //Variables used for population/generation control
    private solution[] population;
    private solution previousBest = new solution();
    public solution bestSolution = new solution();
    private Integer bestSolutionScore;
    private int bestGeneration = 0;
    private int generation = 1;
    private long startTime;

    // Declaring variables used for generation management
    private int maxGenerations = 200;
    private int populationSize = 1000;
    private int tournamentContestants = 3;
    private int percentageOfSelectedPopulationMutated = 10;
    private int percentageOfMutatedSlots = 5;
    private int crossoverType = 3; // 0 = random, 1 = random two point, 2 = fixed mid point, 3 = uniform

    // Toggle functionality variables
    private boolean mutationOn = true;
    private boolean roulette = false;
    private boolean mutationRampingOn = true;
    private boolean universityConstraints = true;

    // Hill climbing toggles/variables
    private int randomHillClimbNumber = 1; // Percentage of random individuals to expose to hill climbing operators
    private int KIndividualsToClimb = 4; // Amount of top individuals to expose to the hill climb operators
    private boolean wholePopHillclimbOn = false;
    private boolean selectedPopHillclimb = false;
    private boolean randomPopHillclimb = true;
    private boolean hillClimbTsOn = false;
    private boolean hillClimbRoomOn = false;

    // Variable used to control mutation ramping
    private int rampingRateInterval = 20;
    private int slotsIncreaseAmount = 5;
    private int amountOfPopMutatedIncrease = 10;
    // Static ramping variables
    private boolean currentlyRamping = false;
    private int rampingCounter = 0;
    private int initialPercentageOfSelectedPopulationMutated = percentageOfSelectedPopulationMutated;
    private int initialPercentageOfMutatedSlots = percentageOfMutatedSlots;

    // End of variable declerations

    public timetable(String inArgs, int counter, boolean testing) throws IOException{
        // Read in data from config file
        runtime = Runtime.getRuntime();
        startTime = System.currentTimeMillis();
        readInput(inArgs);
        checkRunSettings();

        // Generate initial generation
        population = new solution[populationSize];
        initialPop();

        // Run evolution
        for (int i = 1; i < maxGenerations; i++){
            int genFitScore = newGeneration();
            if (!testing){
                long kb = 1024*1024;
                long maxMem = (runtime.totalMemory() - runtime.freeMemory()) / kb;
                System.out.println("Generation: " + i + ",  Fitness Score: " + genFitScore);
                System.out.println("Max mem: " + Long.toString(maxMem) + "mb");
                System.out.println(" ");
            }

            if (genFitScore == 0) break;

            generation += 1;
        }

        // Output final solution
        outputTimetable(inArgs, counter);
    }

    private int newGeneration(){
        /*
        This method creates a new generation by calculating the fitness scores of every solution in
        the population. It then uses the score to create a breeding pool.
        Individuals are selected from that pool to "breed", and also undergo so genetic operations.
         */

        solution[] childPopulation = new solution[populationSize];
        // Assign each solution a fitness score, and save solution with best fitness score
        boolean trigger = false;
        for (solution thisS : population) {
            int score = calcFitnessScore(thisS);
            thisS.setFitnessScore(score);

            // Check if best score
            if (!trigger || previousBest.getFitnessScore() > score){
                previousBest = thisS;
                trigger = true;
            }
        }


        /*
        Tournament selection to create child population.
          */

        for (int i = 1; i < populationSize; i++){
            solution p1 = new solution();
            solution p2 = new solution();
            if (roulette){
                p1.copySolution(population[rouletteSelection()]);
                p2.copySolution(population[rouletteSelection()]);
            }else{
                p1.copySolution(population[tournamentSelection()]);
                p2.copySolution(population[tournamentSelection()]);
            }
            childPopulation[i] = createChild(p1, p2);
        }


        /*
        Create new population. Best solution from previous population added to new generation
         */

        // Perform hill climbs on best K solutions from child population
        if(selectedPopHillclimb){
            int[] topKChildIndexs = topNIndividuals(childPopulation, KIndividualsToClimb);
            for (int i = 0; i < KIndividualsToClimb; i++){
                childPopulation[topKChildIndexs[i]].copySolution(hillClimbTS(childPopulation[topKChildIndexs[i]]));
                childPopulation[topKChildIndexs[i]].copySolution(hillClimbRoom(childPopulation[topKChildIndexs[i]]));

            }
        }

        // Perform hill climbs on random solutions from child population
        if(randomPopHillclimb){
            int randomPercentage = (populationSize * randomHillClimbNumber) / 100;
            int[] topKChildIndexs = new int[randomPercentage];
            for (int i = 0; i < randomPercentage; i++){
                topKChildIndexs[i] = randInRange(0, populationSize - 1);

            }

            for (int i = 0; i < randomPercentage; i++){
                try{
                    childPopulation[topKChildIndexs[i]].copySolution(hillClimbTS(childPopulation[topKChildIndexs[i]]));
                    childPopulation[topKChildIndexs[i]].copySolution(hillClimbRoom(childPopulation[topKChildIndexs[i]]));
                }catch (NullPointerException e){
                    System.out.println(e.getMessage());
                }
            }
        }

        // Perform hill climbs on best solution
        if (hillClimbTsOn){
            previousBest.copySolution(hillClimbTS(previousBest));
        }
        if (hillClimbRoomOn){
            previousBest.copySolution(hillClimbRoom(previousBest));
        }


        population[0] = previousBest;
        for (int i = 1; i < childPopulation.length; i++){
            population[i] = childPopulation[i];
        }

        // PERFORM HILL CLIMB ON WHOLE POPULATION
        if (wholePopHillclimbOn){
            for (int i = 0; i < population.length; i++){
                population[i].copySolution(hillClimbTS(population[i]));
            }
        }

        // Performs mutation ramping is enabled
        if (mutationRampingOn && bestSolutionScore != null) mutationRamping(previousBest.getFitnessScore());

        // Check if best solution so far, and save if it is
        if (bestSolutionScore == null || previousBest.getFitnessScore() < bestSolutionScore){
            bestSolution = previousBest;
            bestSolutionScore = previousBest.getFitnessScore();
            bestGeneration = generation;
        }

        return previousBest.getFitnessScore();
    }

    private solution createChild(solution parent1, solution parent2){
        /*
        Method takes two parent solution and combines them into a child solution
        using these genetic operations:
        . Crossover
        . Mutation
        . Local search
         */

        // CROSSOVER
        solution newChild = crossover(parent1, parent2);

        // MUTATION
        if (randInRange(0,100) < percentageOfSelectedPopulationMutated && mutationOn) {
            newChild = mutateChildOld(newChild);
        }

        return newChild;

    }

    private solution crossover(solution p1, solution p2){
        solution newChild = new solution();
        List<slot> p1Slots = p1.getSlots();
        List<slot> p2Slots = p2.getSlots();
        switch (crossoverType) {
            case 0:
                // Random one point
                int randMidPoint = randInRange(0, p1.getSlots().size());
                //Parent 1
                for (int i = 0; i < randMidPoint; i++){
                    newChild.addSlot(p1Slots.get(i));
                }
                //Parent 2
                for (int i = randMidPoint; i < p2Slots.size(); i++){
                    newChild.addSlot(p2Slots.get(i));
                }
                break;
            case 1:
                // Random two point
                int[] randMidPoints = new int[2];
                randMidPoints[0] = randInRange(0, p1.getSlots().size());
                randMidPoints[1] = randInRange(0, p1.getSlots().size());
                Arrays.sort(randMidPoints);
                // First part of Parent 1
                for (int i = 0; i < randMidPoints[0]; i++){
                    newChild.addSlot(p1Slots.get(i));
                }
                // Part of Parent 2
                for (int i = randMidPoints[0]; i < randMidPoints[1]; i++){
                    newChild.addSlot(p2Slots.get(i));
                }
                // Last part of Parent 1
                for (int i = randMidPoints[1]; i < p1Slots.size(); i++){
                    newChild.addSlot(p1Slots.get(i));
                }
                break;
            case 2:
                // Fixed halfway point
                int halfwayPoint = p1Slots.size() / 2;
                for (int i = 0; i < halfwayPoint; i++){
                    newChild.addSlot(p1Slots.get(i));
                }
                //Parent 2
                for (int i = halfwayPoint; i < p2Slots.size(); i++){
                    newChild.addSlot(p2Slots.get(i));
                }
                break;
            case 3:
                // Uniform
                for (int i = 0; i < p1.getSlots().size(); i++){
                    int randNum = randInRange(0, 1);
                    if (randNum == 0){
                        newChild.addSlot(p1Slots.get(i));
                    }else{
                        newChild.addSlot(p2Slots.get(i));
                    }
                }
                break;
            default:
                System.out.println("Error in crossover: mode 0, 1, 2 not selected correctly.");
                break;
        }
        return newChild;
    }

    private solution mutateChild(solution child){
        /*
        This method take a solution as input and mutates it in the following ways:
        . Mutates 20% of its commitments by selecting a commitment at random and moving the commitment
          the next available time slot
         */

        // IF USED AGAIN, UPDATE TO POST BROKEN FORMAT.... E.G.  GET RID OF .getGenes()


        HashMap<Integer, AtomicInteger> slotSpread = new HashMap<>();
        HashMap<Integer, HashSet<Integer>> slotSpreadRooms = new HashMap<>();

        List<slot> childSlots = child.getSlots();

        // Initiate above variables
        for (int i = 0; i < timeSlots.size(); i++){
            slotSpread.put(i, new AtomicInteger());
            slotSpread.get(i).set(0);
            slotSpreadRooms.put(i, new HashSet<Integer>());
        }

        // Populate with the amount of times a specific timeslot is being used by child, and the room in that timeslot
        for (int i = 0; i < childSlots.size(); i++) {
            slotSpread.get(childSlots.get(i).ts).incrementAndGet();
            slotSpreadRooms.get(childSlots.get(i).ts).add(childSlots.get(i).rm);
        }

        int emptySlot = 0;
        int slotToShift = 0;

        // Find first empty room (first where atomic int = 0)
        for (int i = 0; i < timeSlots.size(); i++){
            if (slotSpread.get(i).intValue() == 0){
                emptySlot = i;
                break;
            }
        }

        // if more than one childslot in a timeslot, and num of rooms =! to instances of that specific slot
        // (therefore indicating a duplicate or clashing timeslot+room combo), then select that as the timeslot to shift
        for (int i = 0; i < timeSlots.size(); i++){
            if(slotSpread.get(i).intValue() > 1){
                if (slotSpreadRooms.get(i).size() != slotSpread.get(i).intValue()){
                    slotToShift = i;
                    break;
                }
            }
        }


        /*
        ERROR BEING CAUSED BY ABOVE FOR LOOP SETTING SLOTTOSHIFT EQUAL TO A TIMESLOT INDEX, NOT A COMMITMENT INDEX.
        THEREFORE IT CAN CAUSE OF OUT BOUNDS + MEANS IT DOESNT WORK AS INTENDED.
         */


        for (int i = 0; i < childSlots.size(); i++){
            int com = childSlots.get(i).com;
            int rm = childSlots.get(i).rm;
            int ts = childSlots.get(i).ts;

            child.getSlots().get(i).setGenes(ts, rm, com);
        }

        child.getSlots().get(slotToShift).setGenes(emptySlot, childSlots.get(slotToShift).rm, childSlots.get(slotToShift).com);

        return child;
    }

    private solution mutateChildOld(solution child){
        /*
        This method take a solution as input and mutates it in the following ways:
        . Mutates 20% of its commitments by selecting a commitment at random and moving the commitment
          the next available time slot
        */

        for (int i = 0; i < child.getSlots().size(); i++){
            if (randInRange(0,100) < percentageOfMutatedSlots){
                int com = child.getSlots().get(i).com;
                int rm;
                int ts;
                if (randInRange(0,2) > 1){
                    rm = randInRange(0, rooms.size() - 1);
                }else{
                    rm = child.getSlots().get(i).com;
                }
                if (randInRange(0,2) > 1){
                    ts = randInRange(0, timeSlots.size() - 1);
                }else{
                    ts = child.getSlots().get(i).com;
                }
                child.getSlots().get(i).setGenes(ts, rm, com);
            }
        }

        return child;

    }

    private solution hillClimbTS(solution child){
        /*
        This method performs local search on a child solution
         */
        try {
            solution newChild = new solution();
            newChild.copySolution(child);
            int bestScore = newChild.getFitnessScore();
            int slotPickAmount = 4;
            int[] randomSlots = new int[slotPickAmount];

            for (int i = 0; i < slotPickAmount; i++) {
                randomSlots[i] = randInRange(0, child.getSlots().size() - 1);
            }

            for (int i = 0; i < randomSlots.length; i++) {
                for (int k = 0; k < timeSlots.size(); k++) {
                    int previousTS = newChild.slots.get(randomSlots[i]).ts;
                    newChild.slots.get(randomSlots[i]).ts = k;
                    int tempScore = calcFitnessScore(newChild);
                    if (tempScore < bestScore) {
                        bestScore = tempScore;
                        break;
                    } else {
                        newChild.slots.get(randomSlots[i]).ts = previousTS;
                    }
                }
            }

            return newChild;
        } catch(NullPointerException e) {
            System.out.println(e.toString());
            return child;
        }

    }

    private solution hillClimbRoom(solution child){
        /*
        This method performs local search on a child solution
         */
        try{
            solution newChild = new solution();
            newChild.copySolution(child);
            int bestScore = newChild.getFitnessScore();
            int slotPickAmount = 4;
            int[] randomSlots = new int[slotPickAmount];


            for (int i = 0; i < slotPickAmount; i++){
                randomSlots[i] = randInRange(0, child.getSlots().size()-1);
            }

            for (int i = 0; i < randomSlots.length; i++){
                for (int k = 0; k < rooms.size(); k++){
                    int previousRoom = newChild.slots.get(randomSlots[i]).rm;
                    newChild.slots.get(randomSlots[i]).rm = k;
                    int tempScore = calcFitnessScore(newChild);
                    if (tempScore < bestScore){
                        bestScore = tempScore;
                        break;
                    }else{
                        newChild.slots.get(randomSlots[i]).rm = previousRoom;
                    }
                }
            }

            return newChild;
        }catch (NullPointerException e){
            System.out.println(e.toString());
            return child;
        }
    }

    private int[] topNIndividuals(solution[] childPop, int kIndividuals){
        // This method returns the top K fitness score individuals from a given population

        List<int[]> solsToSort = new ArrayList<>();
        for (int i = 1; i < childPop.length; i++){
            childPop[i].setFitnessScore(calcFitnessScore(childPop[i]));
            int[] temp = new int[2];
            temp[0] = i;
            temp[1] = childPop[i].getFitnessScore();
            solsToSort.add(temp);
        }

        Collections.sort(solsToSort, new Comparator<int[]>() {
            @Override
            public int compare(int[] o1, int[] o2) {
                return Integer.compare(o1[1], o2[1]);
            }
        });

        int[] toRet = new int[kIndividuals];
        for (int i = 0; i < kIndividuals; i++){
            toRet[i] = solsToSort.get(i)[0];
        }
        return toRet;
    }

    private void mutationRamping(int fitnessScore){
        /*
        This method controls mutation ramping amounts.
         */

        if (fitnessScore == bestSolutionScore){
            if (currentlyRamping){
                if (rampingCounter < rampingRateInterval){
                    rampingCounter += 1;
                }else if (rampingCounter == rampingRateInterval){
                    if (!((percentageOfMutatedSlots + slotsIncreaseAmount) > 100)){
                        percentageOfMutatedSlots += slotsIncreaseAmount;
                    }
                    if (!((percentageOfSelectedPopulationMutated + amountOfPopMutatedIncrease) > 100)){
                        percentageOfSelectedPopulationMutated += amountOfPopMutatedIncrease;
                    }
                    rampingCounter = 0;
                    //System.out.println("Mutation rates have been ramped!");
                }else{
                    System.out.println("Error: rampingCounter > rampingRateInterval");
                }
            }else{
                currentlyRamping = true;
            }
        }else{
            if (currentlyRamping){
                currentlyRamping = false;
                percentageOfMutatedSlots = initialPercentageOfMutatedSlots;
                percentageOfSelectedPopulationMutated = initialPercentageOfSelectedPopulationMutated;
            }
        }
    }

    private int calcFitnessScore(solution s){

        HashMap<Integer, List<slot>> groupedSlots = new HashMap<>();
        int[] blocked = new int[s.getSlots().size()];
        int fitnessScore = 0;

        /*
          Group slots into timeSlots and check the following soft constraints:
          . No 9am Lectures
          . Lectures are in blocks
        */

        for (int i = 0; i < s.getSlots().size(); i++){
            // Groups slots into time slots
            slot tempSlot = s.getOneSlot(i);
            int timeSlot = tempSlot.ts;
            if (groupedSlots.containsKey(timeSlot)){
                groupedSlots.get(timeSlot).add(tempSlot);
            }else{
                groupedSlots.put(timeSlot, new ArrayList<slot>());
                groupedSlots.get(timeSlot).add(tempSlot);
            }

            // Check if time slot is at 9am
            if (timeSlots.get(tempSlot.ts)[1].equals("09:00")){
                fitnessScore += 10;
            }

            // Record all time slots into array and then sort it
            blocked[i] = tempSlot.ts;
        }

        // University specific constraints 14:00 -> 17:00

        if (universityConstraints){

            for (int i = 0; i < s.getSlots().size(); i++){
                // Wednesday after 13:00
                String[] tempTimeSlot = timeSlots.get(s.getOneSlot(i).ts);
                if (tempTimeSlot[0].equals("3") && (tempTimeSlot[1].equals("14:00") || tempTimeSlot[1].equals("15:00") || tempTimeSlot[1].equals("16:00") || tempTimeSlot[1].equals("17:00"))){
                    fitnessScore += 10000;
                }

                // Quiet time
                if (tempTimeSlot[0].equals("5") && tempTimeSlot[1].equals("13:00")){
                    fitnessScore += 10000;
                }

            }


        }

        // Check if Lectures are in blocks
        Arrays.sort(blocked);
        int cap = 0;
        for (int i =0; i < blocked.length; i++){
            if (i != 0){

                if (!((blocked[i] < (blocked[i-1] + 4)) && (timeSlots.get(blocked[i])[0].equals(timeSlots.get(blocked[i-1])[0])))){
                    cap += 0;
                }
            }
        }
        fitnessScore += 10 * cap;

        /*
          For every group of slots in the same timeSlot check these hard constraints:
          . Room collisions
          . Lecturer collisions
          . Student collisions
        */

        for (Integer key : groupedSlots.keySet()){
            // Set of rooms
            HashSet<Integer> roomSet = new HashSet<>();
            // Set of students
            HashSet<String> studentSet = new HashSet<>();
            // Set of Lectures
            HashSet<Integer> lecturerSet = new HashSet<>();
            // For every slot in a timeSlot
            for (int i = 0; i < groupedSlots.get(key).size(); i++){
                slot tempSlot = groupedSlots.get(key).get(i);


                /*for (Integer key2 : tempSlot.getGenes()){
                    System.out.print(key2 + " : ");
                }*/


                // Record slot rooms
                if (!roomSet.add(tempSlot.rm)){
                    fitnessScore += 10000;
                }
                //Record Lecturer clashes
                if (!lecturerSet.add(tempSlot.com)) {
                    fitnessScore += 10000;
                }
                //Record student clashes and add penalty
                List<String> tempList = modules.get(commitments.get(tempSlot.com)[0]);
                for (int j = 0; j < tempList.size(); j++){
                    if (!studentSet.add(tempList.get(j))){
                        fitnessScore += 10000;
                        break;
                    }
                }
            }
        }

        return fitnessScore;
    }

    private void initialPop(){
        /*
            This method is used to generate the initial population of solutions
         */

        // Fill population array with solution instances
        for (int i=0; i < populationSize; i++){
            population[i] = new solution();
        }

        // Make solution instances valid (fill with slots)
        for (int i = 0; i < populationSize; i++){
            for (int j = 0; j < commitments.size(); j++){
                int com = j;
                int ts = randInRange(0, timeSlots.size() - 1);
                int rm = randInRange(0, rooms.size() - 1);

                population[i].addSlot(new slot(ts,rm,com));

            }
        }


    }

    private void outputTimetable(String inputFileName, Integer counter) throws IOException {
        /*
        Method is called after the evolution algorithm has finished. The best final solution
        is then outputted in a human readable way.
         */

        System.out.println();
        System.out.println("Best fitness generated = " + String.valueOf(bestSolution.getFitnessScore()));
        System.out.println("On generation = " + String.valueOf(bestGeneration));


        List<slot> bestSlots = bestSolution.getSlots();

        // Sort slots by days
        Collections.sort(bestSlots, new Comparator<slot>() {
            @Override
            public int compare(slot o1, slot o2) {
                return Integer.compare(o1.ts, o2.ts);
            }
        });

        String currentDay = "";

        FileWriter fw = new FileWriter("/Users/jamesdavis/Google Drive/Dissertation/Main Project/Code/src/files/output_" + inputFileName.substring(0, inputFileName.length()-4) + "_" + counter.toString() + ".txt");
        BufferedWriter bw = new BufferedWriter(fw);

        bw.write("-------------- Fitness Score / Generation --------------");
        bw.newLine();
        bw.write("Fitness score of: " + bestSolution.getFitnessScore() + ", achieved on generation: " + bestGeneration);
        bw.newLine();bw.newLine();

        long kb = 1024;
        long maxMem = runtime.maxMemory() / kb;
        bw.write("-------------- Maximum JVM memory usage --------------");
        bw.newLine();bw.write("Maximum memory = " + Long.toString(maxMem) + "KB");
        bw.newLine();bw.newLine();

        long timeTaken = System.currentTimeMillis() - startTime;
        bw.write("-------------- Time taken to run --------------");
        bw.newLine();bw.write("Time taken = " + Long.toString(timeTaken) + "ms");
        bw.newLine();bw.newLine();

        bw.write("-------------- Violated Constraints --------------");
        int hard = 0;
        int soft = 0;
        int CScore = bestSolutionScore;
        while(CScore != 0){
            if ((CScore - 10000) >= 0){
                hard += 1;
                CScore -= 10000;
            }else if (CScore - 10 >= 0){
                soft += 1;
                CScore -= 10;
            }else{
                break;
            }
        }
        bw.newLine();bw.write("Fitness Score = " + bestSolution.getFitnessScore());
        bw.newLine();bw.write("Hard constraints: " + Integer.toString(hard));bw.newLine();
        bw.write("Soft constraints: " + Integer.toString(soft));

        bw.newLine();bw.newLine();bw.newLine();bw.newLine();

        bw.write("-------------- Timetable --------------");
        bw.newLine();

        for (int i = 0; i < bestSlots.size(); i++){

            String tempRoom = rooms.get(bestSlots.get(i).rm);
            String tempTS[] = timeSlots.get(bestSlots.get(i).ts);
            String tempCom[] = commitments.get(bestSlots.get(i).com);

            if (currentDay.equals("") || !tempTS[0].equals(currentDay)){
                currentDay = tempTS[0];
                bw.newLine();
                bw.write("-------- Day " + currentDay + " --------");
                bw.newLine();
            }

            bw.write("Time: " + tempTS[1] + ", Room: " + tempRoom +
                    ", Module: " + tempCom[0] + ", Lecturer: " + tempCom[1]);
            bw.newLine();

        }

        bw.close();
        fw.close();

    }

    private void readInput(String fileName) throws IOException {
        /*
            Method takes the filename of the config file as input and loads
            the data from the file into the corresponding places.
         */

        File f = new File(fileName);
        BufferedReader b = new BufferedReader(new FileReader(f));
        String readLine = "";
        Boolean started = false;
        Boolean end = false;
        Boolean tsFlag = false;
        Boolean rFlag = false;
        Boolean cFlag = false;
        Boolean mFlag = false;

        while (((readLine = b.readLine()) != null) || !end) {
            //System.out.println(readLine);
            if (readLine.equalsIgnoreCase("::start::")){
                started = true;
            }else if (started){
                /*
                Main body of code to look for new flags.
                 */

                if (readLine.equalsIgnoreCase("::ts::")){
                    if (tsFlag){
                        tsFlag = false;
                    }else{
                        tsFlag = true;
                    }
                }else if (readLine.equalsIgnoreCase("::r::")){
                    if (rFlag){
                        rFlag = false;
                    }else{
                        rFlag = true;
                    }
                }else if (readLine.equalsIgnoreCase("::c::")){
                    if (cFlag){
                        cFlag = false;
                    }else{
                        cFlag = true;
                    }
                }else if (readLine.equalsIgnoreCase("::m::")){
                    if (mFlag){
                        mFlag = false;
                    }else{
                        mFlag = true;
                    }
                }else if (readLine.equalsIgnoreCase("::end::")){
                    end = true;
                }else if (readLine.substring(0, 1).equalsIgnoreCase("[") || rFlag){

                    // Reads in contents of a specific flag
                    if (tsFlag){
                        readLine = readLine.substring(1, readLine.length() - 1);
                        String[] contents = readLine.split(",");
                        timeSlots.add(contents);
                    }else if (rFlag){
                        rooms.add(readLine.replaceAll("\\s+", ""));
                    }else if (cFlag){
                        readLine = readLine.substring(1, readLine.length() - 1);
                        String[] contents = readLine.split(",");
                        commitments.add(contents);
                    }else if (mFlag){
                        readLine = readLine.substring(1, readLine.length() - 1);
                        String[] contents = readLine.split(",");
                        if (modules.containsKey(contents[0])){
                            modules.get(contents[0]).add(contents[1]);
                        }else{
                            modules.put(contents[0], new ArrayList<String>());
                            modules.get(contents[0]).add(contents[1]);
                        }
                    }else{
                        System.out.println("Started and no flags read/activated");
                    }
                }else{
                    System.out.println("Invalid line: doesn't start with flag or [");
                }
            }
            else {
                System.out.println("No start flag found - ::start::");
            }
        }
    }

    private int rouletteSelection(){
        // This method selects a solution from the population using roulette wheel selection

        // Find largest fitness score present in population
        int maxScore = 0;
        for (int i = 0; i < populationSize; i++){
            if (population[i].getFitnessScore() > maxScore){
                maxScore = population[i].getFitnessScore();
            }
        }

        /*
        For each individual, do maxScore - fitness score. Add 10 to stop max fitness
        individuals being removed from selection
         */
        int[] invertedScores = new int[populationSize];
        for (int i = 0; i < populationSize; i++){
            invertedScores[i] = (maxScore - population[i].getFitnessScore()) + 10;
        }

        // Sum all the inverted scores
        int sum = 0;
        for(int i = 0; i < populationSize; i++){
            sum += invertedScores[i];
        }

        // Create selection weighting
        double[] weighting = new double[populationSize];
        for (int i = 0; i < populationSize; i++){
            weighting[i] = (double)invertedScores[i] / (double)sum;
        }

        // Select individual
        Random r = new Random();
        double randomVal = Math.random();//r.nextDouble();
        int index = 0;
        while (true){
            randomVal = randomVal - weighting[index];
            if (randomVal <= 0){
                return index;
            }else{
                index += 1;
            }
        }
    }

    private int tournamentSelection(){

        /*
           Method randomly selects solutions from the population and returns the
           index of the solution with the lowest fitness score
        */

        int best = -1;
        for (int i = 0; i < tournamentContestants; i++){
            int randNum = randInRange(0, populationSize - 1);
            int fScore = population[randNum].getFitnessScore();

            if (best == -1 || fScore < population[best].getFitnessScore()){
                best = randNum;
            }
        }
        return best;
    }

    private void checkRunSettings(){
        int total = 0;
        if (wholePopHillclimbOn){total++;}
        if (selectedPopHillclimb){total++;}
        if (randomPopHillclimb){total++;}
        if (hillClimbTsOn || hillClimbRoomOn){total++;}
        if (total > 1){
            System.out.println("Warning!: More than one hillclimb setting enabled, expect slow generation.");
        }
    }

    private int randInRange(int min, int max){
        // E.g. randInRange(0,1) will randomly produce either a 0 or a 1
        int range = (max - min) + 1;
        return (int)(Math.random() * range) + min;
    }
}
